#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MaxQuintuples     128 /* expand as needed */
#define QuintupleLen        5
#define MaxBuffer         128
#define MaxTape            33 /* expand as needed: 1 line of display */
#define MaxSteps      1000000 /* assume 'infinite looping' thereafter */
#define Blank            '0'  /* 2-character alphabet: 0 and 1 */
#define StartState       'a'

#define TapeBorder       "+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+"
#define CellSide         '|'
#define Ellipsis         "..."
#define CellLen          2   

enum { NewState = 2, NewSymbol = 3, Direction = 4 }; 
char quintuples[MaxQuintuples][QuintupleLen + 1]; /* array of strings */
unsigned qlen = 0;        /* number of entries from input file */
unsigned displayFlag = 1; /* 2nd command-line arg turns this off */

char tape[MaxTape];
unsigned currentCell = MaxTape / 2; /* tape middle */
char currentState = StartState;
unsigned instructionsExecuted = 0;

void die(const char* msg) {
  puts(msg);
  exit(EXIT_FAILURE);
}

void print_cell(unsigned ind) {
  putchar(CellSide);
  putchar(tape[ind]);
}

void pause() {
  puts("Hit RETURN to continue...");
  getchar();
}

void display_tape() {
  printf("\nCurrent state: %c\n", currentState);
  printf("   %s\n", TapeBorder);
  printf("%s", Ellipsis);
  
  unsigned i;
  for (i = 0; i < MaxTape; i++) print_cell(i);
  printf("%c%s\n   %s\n", CellSide, Ellipsis, TapeBorder);

  i = (CellLen * currentCell) + 2;
  printf("%*s%c\n", i, "", '^');
  pause();
}

int comp(const void* e1, const void* e2) { /* qsort callback */
  const char* s1 = (char*) e1;
  const char* s2 = (char*) e2;
  return strncmp(s1, s2, 2); /* < 0 = s1 < s2; 0 = s1 == s2; > 0 = s1 > s2 */
}

void read_program(const char* file) {
  FILE* infile = fopen(file, "r");
  char buff[MaxBuffer + 1];
  if (NULL == infile) {
    sprintf(buff, "Can't open program file %s", file);
    die(buff);
  }
   
  while (fgets(buff, MaxBuffer, infile)) { /* read until end-of-file */
    if ('#' == buff[0]) continue;              /* ignore comments */
    if (strlen(buff) < QuintupleLen) continue; /* ignore faulty lines */
    strncpy(quintuples[qlen], buff, QuintupleLen);
    qlen++;
  }
  fclose(infile);

  qsort(quintuples, qlen, sizeof(quintuples[0]), comp); /* sort for easy access */
  memset(tape, Blank, MaxTape); /* blank out the tape */
}

void report() {
  /* Show instructions. */
  printf("%i instructions:\n", qlen);
  unsigned i, count = 0;
  for (i = 0; i < qlen; i++) puts(quintuples[i]);
  for (i = 0; i < MaxTape; i++) if ('1' == tape[i]) count++;
  printf("Total 1s on tape:      %8i\n", count);
  printf("Instructions executed: %8i\n", instructionsExecuted);
}

void check_for_errors(const char* action) {
  if (0 == (currentCell - 1) && 'L' == action[Direction]) die("Can't move left...");
  if (currentCell >= MaxTape - 1 && 'R' == action[Direction]) die("Can't move right...");
  if (instructionsExecuted >= MaxSteps) die("Seems to be infinitely touring...");
}

void run_simulation() {
  while (1) {
     if (displayFlag) display_tape();
  
    /* Get the action for the current key. */
    char key[3];
    sprintf(key, "%c%c", currentState, tape[currentCell - 1]);
    char* action = bsearch(key, quintuples, qlen, sizeof(quintuples[0]), comp);
    if (NULL == action) break; /* no match == normal termination */

    check_for_errors(action);

    /* Update system. */
    currentState = action[NewState];
    tape[currentCell - 1] = action[NewSymbol];
    if ('L' == action[Direction]) currentCell--; /* move left */
    else currentCell++;                          /* move right */
    instructionsExecuted++;                      /* update step counter */
  }
}

int main(int argc, char* argv[]) {
  if (argc < 2) die("Usage: turing <program file>");
  if (argc > 2 && 0 == strcmp(argv[2], "off")) displayFlag = 0;
  
  read_program(argv[1]);
  run_simulation();
  report();
  return 0;
}

